// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
/***************************************************************
*
* DataSetStruct.cc	& XMM Science Analysis System (SAS)
*
* Class representing abstract data type DataSetStruct
*
* (c) 2000 ESA - Written by Mark Thomas, CARA
*
* $Id: DataSetStruct.cc,v 1.6 2009/09/01 16:13:55 aibarra Exp $
*
*****************************************************************/


#include <cstdarg>
#include <map>
#include <errstr.h>
#include <Tracer.h>
#include <scanner.h>
#include <dal_utilities.h>
#include <DalChild.h>


#include "DataSetStruct.h"


static const string		datasetDirective = "dataset";
static const string		tableDirective = "table";
static const string		arrayDirective = "array";
static const string		columnDirective = "column";
static const string		includeDirective = "include";
static const string		nameDirective = "name";
static const string		attributeDirective = "attribute";
static const string		typeDirective = "type";
static const string		valueDirective = "value";
static const string		rowsDirective = "rows";


static void
expectedOneOf( ScannerT & scanner, const char * string1, ... )
{
	PROC
	va_list		args;
	char *		arg;
	string		strings( string1 );

	va_start( args, string1 );

	while( (arg = va_arg( args, char * )) != 0 )
	{
		strings += string( ", " ) + arg;
	} // while

	va_end( args );

	StreamT & file = scanner.stream();

	errstr << file.FileNameOf() << ": Error at line " <<
		file.CurrentLineNo() <<
 		", and at or near column " <<
		file.CurrentColumnNo() <<
		" - Expected one of " << strings << Sas::fatal("expected one of");

	ENDPROC
}

/*
static void
checkAttribute(
	const Attributable*	att,
	const AttributeStruct*	attributeStruct )
{
	PROC
	if( !att -> hasAttribute( attributeStruct -> name() ) )
	{
		errstr << "Expected attributable with qualified name '" << att -> qualifiedName() << "' to have an attribute with name '" << attributeStruct -> name() << "'" << Sas::error( "attributeNotFound" );
	} // if

	const Attribute* checkThisAttribute = att -> attribute( attributeStruct -> name() );

	if( checkThisAttribute -> dataType() != attributeStruct -> dataType() )
	{
		errstr << "Expected attribute with qualified name '" << checkThisAttribute -> qualifiedName() << "' to have data type " << toString( attributeStruct -> dataType() ) << " but found " << toString( checkThisAttribute -> dataType() ) << Sas::error( "wrongAttributeDataType" );
	} // if

	if( checkThisAttribute -> asString() != attributeStruct -> stringValue() )
	{
		errstr << "The attribute with qualified name '" << checkThisAttribute -> qualifiedName() << "' was expected to have value " << attributeStruct -> stringValue() << " but found " << checkThisAttribute -> asString() << Sas::error( "wrongAttributeValue" );
	} // if
	ENDPROC
}


static void
checkAttributable(
	const Attributable*		att,
	const AttributableStruct*	attributableStruct )
{
	PROC
	if( att -> attributes() != attributableStruct -> attributeStructs() )
	{
		errstr << "Expected attributable with qualified name '" << att -> qualifiedName() << "' to have " << attributableStruct -> attributeStructs() << " attributes, but " << att -> attributes() << " were found" << Sas::error( "wrongNumberOfAttributes" );
	} // if

	for(
		unsigned int i = 0;
		i < attributableStruct -> attributeStructs();
		++i )
	{
		checkAttribute( att, attributableStruct -> attributeStruct( i ) );
	} // for
	ENDPROC
}


static void
checkColumn(
	const Table*		tb,
	const ColumnStruct*	columnStruct )
{
	PROC
	if( !tb -> hasColumn( columnStruct -> name() ))
	{
		errstr << "The table with qualified name '" << tb -> qualifiedName() << "' was expected to contain a column with name '" << columnStruct -> name() << "'" << Sas::error( "missingColumn" );
	} // if

	const Column* cl = tb -> column( columnStruct -> name() );

	checkAttributable( cl, columnStruct );

	if( columnStruct -> dataType() != cl -> dataType() )
	{
		errstr << "Expected column with qualified name '" << cl -> qualifiedName() << "' to have data type " << toString( columnStruct -> dataType() ) << " but found " << toString( cl -> dataType() ) << Sas::error( "wrongColumnDataType" );
	} // if
	ENDPROC
}


static void
checkTable(
	const DataSet*		ds,
	const TableStruct*	tableStruct )
{
	PROC
	const Block* bk = ds -> block( tableStruct -> name() );

	checkAttributable( bk, tableStruct );

	if( bk -> type() != Block::TableT )
	{
		errstr << "Expected block with qualifed name '" << bk -> qualifiedName() << "' to be a table" << Sas::error( "wrongBlockType" );
	} // if

	const Table* tb = ds -> table( tableStruct -> name() );

	if( tableStruct -> columnStructs() != tb -> columns() )
	{
		errstr << "Expected table with qualified name '" << tb -> qualifiedName() << "' to contain " << tableStruct -> columnStructs() << " columns but found " << tb -> columns() << Sas::error( "wrongNumberOfColumns" );
	} // if

	for(
		unsigned int i = 0;
		i < tableStruct -> columnStructs();
		++i )
	{
		checkColumn( tb, tableStruct -> columnStruct( i ) );
	} // for
	ENDPROC
}


static void
checkArray(
	const DataSet*		ds,
	const ArrayStruct*	arrayStruct )
{
	PROC

	const Block* bk = ds -> block( arrayStruct -> name() );

	checkAttributable( bk, arrayStruct );

	if( bk -> type() != Block::ArrayT )
	{
		errstr << "Expected block with qualified name '" << bk -> qualifiedName() << "' to be an array" << Sas::error( "wrongBlockType" );
	} // if

	const Array* ar = ds -> array( arrayStruct -> name() );

	if( arrayStruct -> dataType() != ar -> dataType() )
	{
		errstr << "Expected array with qualified name '" << ar -> qualifiedName() << "' to have data type " << toString( arrayStruct -> dataType() ) << " but found " << toString( ar -> dataType() ) << Sas::error( "wrongArrayDataType" );
	} // if
	ENDPROC
}


static void
checkBlock(
	const DataSet*		ds,
	const BlockStruct*	blockStruct )
{
	PROC

	if( !ds -> hasBlock( blockStruct -> name() ))
	{
		errstr << "Expected dataset with name '" << ds -> name() << "' to contain a block with name " << blockStruct -> name() << Sas::error( "missingBlock" );
	} // if

	switch( blockStruct -> type() )
	{
		case Block::TableT :
			checkTable( ds, blockStruct -> parent() -> tableStruct( blockStruct -> number() ));
			break;
		case Block::ArrayT :
			checkArray( ds, blockStruct -> parent() -> arrayStruct( blockStruct -> number()) );
			break;
		default :
			errstr << Sas::error( "internalError" );
			break;
	} // switch
	ENDPROC
}


static void
checkDataSet( const DataSetStruct* dataSetStruct )
{
	PROC
	if( !dataSetServer -> exists( dataSetStruct -> name() ) )
	{
		errstr << "Expected a dataset with name '" << dataSetStruct -> name() << "' to exist" << Sas::error( "datasetDoesNotExist" );
	} // if

	DataSet* ds = dataSetServer -> open( dataSetStruct -> name(), Dal::Read );

	checkAttributable( ds, dataSetStruct );

	if( ds -> blocks() != dataSetStruct -> blockStructs() )
	{
		errstr << "Set " << ds -> name() << " contains " << ds -> blocks();
		errstr << " but expected " << dataSetStruct -> blockStructs();
		errstr << " blocks." << Sas::error( "wrongNumberOfBlocks" );
	} // if

	for(
		unsigned int i = 0;
		i < dataSetStruct -> blockStructs();
		++i )
	{
		checkBlock( ds, dataSetStruct -> blockStruct( i ) );
	} // for
	dataSetServer -> close( ds );

	ENDPROC
}
*/


// *********************
// Forward Declarations.
// *********************
class DalAttributeStruct;
class DalColumnStruct;
class DalDataSetStruct;


// ****************************
// Checkable class beyond here.
// ****************************

class Checkable
{
	public :
		Checkable( bool stopIfCheckFails );
		bool stopIfCheckFails() const;
		//void stopIfCheckFails( bool stopIfCheckFails );

	private :
		bool	_stopIfCheckFails;
};


Checkable::Checkable( bool stopIfCheckFails )
	: _stopIfCheckFails( stopIfCheckFails )
{
	PROC
	ENDPROC
}


bool
Checkable::stopIfCheckFails() const
{
	PROC
	RETURN _stopIfCheckFails;
}


//void
//Checkable::stopIfCheckFails( bool stopIfCheckFails )
//{
//	PROC
//	_stopIfCheckFails = stopIfCheckFails;
//	ENDPROC
//}


// ***********************************
// DalLabelledStruct class beyond here
// ***********************************

class DalLabelledStruct : public virtual LabelledStruct
{
	public :
		DalLabelledStruct( unsigned int number );
		~DalLabelledStruct();

		const string&		name() const;
		void			name( const string& name );
		bool			nameFound() const;
		unsigned int		number() const;

		void			readNameStruct( ScannerT& scanner );

	private :
		DalLabelledStruct( const DalLabelledStruct& );
		DalLabelledStruct& operator=( const DalLabelledStruct& );

		string			_name;
		bool			_nameFound;
		unsigned int		_number;
};


DalLabelledStruct::DalLabelledStruct( unsigned int number )
	: _nameFound( false ),
	  _number ( number )
{
	PROC
	ENDPROC
}

DalLabelledStruct::~DalLabelledStruct()
{
	PROC
	ENDPROC
}


const string&
DalLabelledStruct::name() const
{
	PROC
	RETURN _name;
}


void
DalLabelledStruct::name( const string& name )
{
	PROC
	_name = name;
	ENDPROC
}


bool
DalLabelledStruct::nameFound() const
{
	PROC
	RETURN _nameFound;
}


unsigned int
DalLabelledStruct::number() const
{
	PROC
	RETURN _number;
}


void
DalLabelledStruct::readNameStruct( ScannerT& scanner )
{
	PROC
	Token	token;
	scanner.SetState( ScannerT::LINE );
	scanner.match( nameDirective );
	scanner.Match( Token::String, token );
	_name = token.asString();
	scanner.SetState( ScannerT::NORMAL );
	_nameFound = true;
	ENDPROC
}


// ***************************************
// DalAttributableStruct class beyond here
// ***************************************

class DalAttributableStruct : public virtual AttributableStruct, public DalLabelledStruct, public Checkable
{
	public :
		DalAttributableStruct( unsigned int number, bool stopIfCheckFails );
		~DalAttributableStruct();

		unsigned long		attributeStructs() const;
		const AttributeStruct*	attributeStruct( unsigned int attributeNumber ) const;

		void 			readAttributeStruct( ScannerT& scanner );
		void			check( const Attributable* attributable ) const;

	private :
		// Forbid
		DalAttributableStruct( const DalAttributableStruct& );
		DalAttributableStruct& operator=( const DalAttributableStruct& );

		//typedef map<string,AttributeStruct*>	AttributeList;
		typedef vector<DalAttributeStruct*>	AttributeList;
		AttributeList		_attributes;
};


// ************************************
// DalAttributeStruct class beyond here
// ************************************

class DalAttributeStruct : public virtual AttributeStruct, public virtual DalLabelledStruct, public virtual DalChild<AttributableStruct>, public Checkable
{
	public :
		DalAttributeStruct( unsigned int number, DalAttributableStruct* parent );
		~DalAttributeStruct();

		Attribute::DataType	dataType() const;
		int			intValue() const;
		double			realValue() const;
		string			stringValue() const;
		bool			boolValue() const;
		bool			dataTypeFound() const;
		bool			valueFound() const;

		void 			readAttributeStruct( ScannerT& scanner );
		void			check( const Attributable* attributable ) const;

	private :
		// Forbid
		DalAttributeStruct( const DalAttributeStruct& );
		DalAttributeStruct& operator=( const DalAttributeStruct& );

		Attribute::DataType	_dataType;
		//class Value;
		//Value*			_value;
		Token			_value;
		bool			_dataTypeFound;
		bool			_valueFound;
		void			check( const Attribute* attribute ) const;
};

/*
class DalAttributeStruct::Value
{
	public :

		virtual ~Value() {}
		virtual int		intValue() const = 0;
		virtual double		realValue() const = 0;
		virtual const string&	stringValue() const = 0;
		virtual bool		boolValue() const = 0;

};


class IntValue : public virtual DalAttributeStruct::Value
{
	public :
		int	intValue() const { PROC RETURN _value; }
	private :
		int	_value;
};


class RealValue : public virtual DalAttributeStruct::Value
{
	public :
		double	realValue() const { PROC RETURN _value; }
	private :
		double	_value;
};


class StringValue : public virtual DalAttributeStruct::Value
{
	public :
		const string&	stringValue() const { PROC RETURN _value; }
	private :
		string	_value;
};


class BoolValue : public virtual DalAttributeStruct::Value
{
	public :
		bool	boolValue() const { PROC RETURN _value; }
	private :
		bool	_value;
};
*/


DalAttributeStruct::DalAttributeStruct( unsigned int number, DalAttributableStruct* parent )
	: DalLabelledStruct( number ),
	  DalChild<AttributableStruct>( parent ),
	  Checkable( parent -> stopIfCheckFails() ),
	  _dataTypeFound( false ), _valueFound( false )

{
	PROC
	ENDPROC
}


DalAttributeStruct::~DalAttributeStruct()
{
	PROC
	//delete _value;
	ENDPROC
}


Attribute::DataType
DalAttributeStruct::dataType() const
{
	PROC
	RETURN _dataType;
}


int
DalAttributeStruct::intValue() const
{
	PROC
	//RETURN _value -> intValue();
	RETURN _value.asInt();
}


double
DalAttributeStruct::realValue() const
{
	PROC
	RETURN _value.asReal();
}


string
DalAttributeStruct::stringValue() const
{
	PROC
	RETURN _value.asString();
}


bool
DalAttributeStruct::boolValue() const
{
	PROC
	RETURN _value.asBool();
}


bool
DalAttributeStruct::dataTypeFound() const
{
	PROC
	RETURN _dataTypeFound;
}


bool
DalAttributeStruct::valueFound() const
{
	PROC
	RETURN _valueFound;
}


void
DalAttributeStruct::readAttributeStruct( ScannerT& scanner )
{
	PROC

	scanner.match( attributeDirective );
	scanner.match( '<' );

	while( !scanner.NextIs( '>' ) && !scanner.EndOfFile() )
	{
		if( scanner.nextIs( nameDirective ))
		{
			readNameStruct( scanner );
		}
		else if( scanner.nextIs( typeDirective ))
		{
			Token		token;
			scanner.SetState( ScannerT::LINE );
			scanner.match( typeDirective );
			/* AI gcc4 porting
			 * error: deprecated conversion from string constant to 'char*'
			 * cast to (char *)
			 */
			token = scanner.MatchOneOf( (char *) "Int", (char *)"Real", (char *) "Bool", (char *) "String", 0 );
			_dataType = toAttributeDataType( token.asString() );
			scanner.SetState( ScannerT::NORMAL );
		}
		else if( scanner.nextIs( valueDirective ))
		{
			//parseValueDirective( scanner, newAttribute -> _value );
			//newAttribute -> _valueFound = true;
			Token	token;
			scanner.SetState( ScannerT::LINE );
			scanner.match( valueDirective );
			_value = scanner.acceptNext();
			scanner.SetState( ScannerT::NORMAL );
			//switch( token -> id() )
			//{
			//	case Token::Integer :
			//		_value = new AttributeStruct::Value;
			//		break;
			//	case Token::Real :
			//		_value = new AttributeStruct::Value;
			//		break;
			//	case Token::String :
			//		_value = new AttributeStruct::Value;
			//		break;
			//	case Token::Id :
			////		_value = new AttributeStruct::Value;
			//		break;
			//	default :
			//		errstr << "Type of attribute value of attribute is unrecognised" << Sas:error( "unrecognisedValueType" );
			//} // switch
		}
		else
		{
			expectedOneOf(
				scanner,
				nameDirective.c_str(),
				typeDirective.c_str(),
				valueDirective.c_str(),
				0 );
		} // if
	} // while

	scanner.match( '>' );
	ENDPROC
}


void
DalAttributeStruct::check( const Attribute* attribute ) const
{
	PROC
	if( attribute -> asString() != stringValue() )
	{
		errstr << "The attribute with qualified name '" << attribute -> qualifiedName() << "' was expected to have value " << stringValue() << " but found " << attribute -> asString();

		if( stopIfCheckFails() )
		{
			errstr << Sas::error( "wrongAttributeValue" );
		}
		else
		{
			errstr << Sas::warning( "wrongAttributeValue" );
		}// if
	} // if
	ENDPROC
}


void
DalAttributeStruct::check( const Attributable*	att ) const
{
	PROC
	if( !att -> hasAttribute( name() ) )
	{
		errstr << "Expected attributable with qualified name '" << att -> qualifiedName() << "' to have an attribute with name '" << name() << "'" << Sas::error( "attributeNotFound" );
	} // if

	const Attribute* checkThisAttribute = att -> attribute( name() );

	if( checkThisAttribute -> dataType() != dataType() )
	{
		errstr << "Expected attribute with qualified name '" << checkThisAttribute -> qualifiedName() << "' to have data type " << toString( dataType() ) << " but found " << toString( checkThisAttribute -> dataType() ) << Sas::error( "wrongAttributeDataType" );
	} // if

	check( checkThisAttribute );
	ENDPROC
}


// ***************************************
// DalAttributableStruct class beyond here
// ***************************************

DalAttributableStruct::DalAttributableStruct( unsigned int number, bool stopIfCheckFails )
	:
		DalLabelledStruct( number ),
		Checkable( stopIfCheckFails )
{
	PROC
	ENDPROC
}


DalAttributableStruct::~DalAttributableStruct()
{
	PROC
	ENDPROC
}


unsigned long
DalAttributableStruct::attributeStructs() const
{
	PROC
	RETURN _attributes.size();
}


const AttributeStruct*
DalAttributableStruct::attributeStruct( unsigned int attributeNumber ) const
{
	PROC
	RETURN _attributes[ attributeNumber ];
}

void
DalAttributableStruct::readAttributeStruct( ScannerT& scanner )
{
	PROC

	DalAttributeStruct* dalAttributeStruct = new DalAttributeStruct( _attributes.size(), this );
	dalAttributeStruct -> readAttributeStruct( scanner );
	_attributes.push_back( dalAttributeStruct );

	ENDPROC
}


void
DalAttributableStruct::check( const Attributable*		att ) const
{
	PROC
	if( att -> attributes() != attributeStructs() )
	{
		errstr << "Expected attributable with qualified name '" << att -> qualifiedName() << "' to have " << attributeStructs() << " attributes, but " << att -> attributes() << " were found" << Sas::error( "wrongNumberOfAttributes" );
	} // if

	for(
		unsigned int i = 0;
		i < attributeStructs();
		++i )
	{
		_attributes[ i ] -> check( att );
	} // for
	ENDPROC
}


// *****************************
// DalBlockStruct class beyond here
// *****************************

class DalBlockStruct : public virtual BlockStruct, public DalAttributableStruct, public DalChild<DataSetStruct>
{
	public :
		DalBlockStruct( unsigned int number, DalDataSetStruct* parent );
		~DalBlockStruct();

		virtual void check( const DataSet* dataSet ) const = 0;

	private :
		// Forbid
		DalBlockStruct( const DalBlockStruct& );
		DalBlockStruct& operator=( const DalBlockStruct& );
};


// ********************************
// DalTableStruct class beyond here
// ********************************

class DalTableStruct : public virtual TableStruct, public DalBlockStruct
{
	public :
		DalTableStruct( unsigned int number, DalDataSetStruct* parent );
		~DalTableStruct();

		Block::Type		type() const;
		unsigned int		columnStructs() const;
		const ColumnStruct*	columnStruct( unsigned int columnNumber ) const;
		unsigned long		rows() const;
		bool			rowsFound() const;

		void 			readTableStruct( ScannerT& scanner );
		void			check( const DataSet* dataSet ) const;

	private :
		// Forbid
		DalTableStruct( const DalTableStruct& );
		DalTableStruct& operator=( const DalTableStruct& );

		typedef vector<DalColumnStruct*>	ColumnList;
		ColumnList		_columns;
		unsigned long		_rows;
		bool			_rowsFound;
};


// *********************************
// DalColumnStruct class beyond here
// *********************************

class DalColumnStruct : public virtual ColumnStruct, public DalAttributableStruct, public DalChild<TableStruct>
{
	public :
		DalColumnStruct( unsigned int number, DalTableStruct* parent );
		~DalColumnStruct();

		Column::DataType	dataType() const;
		bool			dataTypeFound() const;

		void 			readColumnStruct( ScannerT& scanner );
		void			check( const Table* tab ) const;

	private :
		// Forbid
		DalColumnStruct( const DalColumnStruct& );
		DalColumnStruct& operator=( const DalColumnStruct& );

		Column::DataType	_dataType;
		bool			_dataTypeFound;
};


DalColumnStruct::DalColumnStruct( unsigned int number, DalTableStruct* parent )
	: DalAttributableStruct( number, parent -> stopIfCheckFails() ),
	  DalChild<TableStruct>( parent ),
	  _dataTypeFound( false )
{
	PROC
	ENDPROC
}


DalColumnStruct::~DalColumnStruct()
{
	PROC
	ENDPROC
}


Column::DataType
DalColumnStruct::dataType() const
{
	PROC
	RETURN _dataType;
}


bool
DalColumnStruct::dataTypeFound() const
{
	PROC
	RETURN _dataTypeFound;
}


void
DalColumnStruct::readColumnStruct( ScannerT& scanner )
{
	PROC
	scanner.match( columnDirective );
	scanner.match( '<' );

	while( !scanner.NextIs( '>' ) && !scanner.EndOfFile() )
	{
		if( scanner.nextIs( nameDirective ))
		{
			readNameStruct( scanner );
		}
		else if( scanner.nextIs( attributeDirective ))
		{
			readAttributeStruct( scanner );
		}
		else if( scanner.nextIs( typeDirective ))
		{
			Token	token;
			scanner.SetState( ScannerT::LINE );
			scanner.match( typeDirective );
			/* AI gcc4 porting
			 * error: deprecated conversion from string constant to 'char*'
			 * cast to (char *)
			 */
			token = scanner.MatchOneOf((char *) "Bool", (char *)"Int8", (char *)"Uint16", (char *)"Int16", (char *)"Uint32", (char *)"Int32", (char *)"Real32", (char *)"Real64", (char *)"String", 0 );
			_dataType = toDalDataType( token.asString() );
			scanner.SetState( ScannerT::NORMAL );
		}
		else
		{
			expectedOneOf(
				scanner,
				nameDirective.c_str(),
				typeDirective.c_str(),
				attributeDirective.c_str(),
				0 );
		} // if
	} // while
	scanner.match( '>' );
	ENDPROC
}


void
DalColumnStruct::check( const Table*		tb ) const
{
	PROC
	if( !tb -> hasColumn( name() ))
	{
		errstr << "The table with qualified name '" << tb -> qualifiedName() << "' was expected to contain a column with name '" << name() << "'" << Sas::error( "missingColumn" );
	} // if

	const Column* cl = tb -> column( name() );

	this -> DalAttributableStruct::check( cl );

	if( dataType() != cl -> dataType() )
	{
		errstr << "Expected column with qualified name '" << cl -> qualifiedName() << "' to have data type " << toString( dataType() ) << " but found " << toString( cl -> dataType() ) << Sas::error( "wrongColumnDataType" );
	} // if
	ENDPROC
}


// **********************************
// DalDataSetStruct class beyond here
// **********************************

class DalDataSetStruct : public virtual DataSetStruct, public DalAttributableStruct, public DalChild<DataSetStructServer>
{
	public :
		DalDataSetStruct( unsigned int number, DataSetStructServer* parent, bool stopIfCheckFails );
		~DalDataSetStruct();

		unsigned int 		blockStructs() const;
		const BlockStruct* 	blockStruct( unsigned int blockNumber ) const;
		const ArrayStruct* 	arrayStruct( unsigned int blockNumber ) const;
		const TableStruct* 	tableStruct( unsigned int blockNumber ) const;
		void			check() const;

		void 			readDataSetStruct( ScannerT& scanner );

	private :
		// Forbid
		DalDataSetStruct( const DalDataSetStruct& );
		DalDataSetStruct& operator=( const DalDataSetStruct& );

		typedef vector<DalBlockStruct*>	BlockList;
		BlockList		_blocks;

};


// *****************************
// DalBlockStruct class beyond here
// *****************************

DalBlockStruct::DalBlockStruct( unsigned int number, DalDataSetStruct* parent )
	: DalAttributableStruct( number, parent -> stopIfCheckFails() ),
	  DalChild<DataSetStruct>( parent )
{
	PROC
	ENDPROC
}


DalBlockStruct::~DalBlockStruct()
{
	PROC
	ENDPROC
}


// ********************************
// DalTableStruct class beyond here
// ********************************

DalTableStruct::DalTableStruct( unsigned int number, DalDataSetStruct* parent )
	: DalBlockStruct( number, parent ),
	  _rowsFound( false )
{
	PROC
	ENDPROC
}


DalTableStruct::~DalTableStruct()
{
	PROC
	ENDPROC
}


Block::Type
DalTableStruct::type() const
{
	PROC
	RETURN Block::TableT;
}


unsigned int
DalTableStruct::columnStructs() const
{
	PROC
	RETURN _columns.size();
}


const ColumnStruct*
DalTableStruct::columnStruct( unsigned int columnNumber ) const
{
	PROC
	RETURN _columns[ columnNumber ];
}


unsigned long
DalTableStruct::rows() const
{
	PROC
	RETURN _rows;
}


bool
DalTableStruct::rowsFound() const
{
	PROC
	RETURN _rowsFound;
}


void
DalTableStruct::readTableStruct( ScannerT& scanner )
{
	PROC
	scanner.match( tableDirective );
	scanner.match( '<' );
	Token	token;

	while( !scanner.NextIs( '>' ) && !scanner.EndOfFile() )
	{
		if( scanner.nextIs( nameDirective ))
		{
			readNameStruct( scanner );
		}
		else if( scanner.nextIs( rowsDirective ))
		{
			scanner.SetState( ScannerT::LINE );
			scanner.match( rowsDirective );
			scanner.Match( Token::Integer, token );
			_rows = token.asInt();
			scanner.SetState( ScannerT::NORMAL );
		}
		else if( scanner.nextIs( attributeDirective ))
		{
			readAttributeStruct( scanner );
		}
		else if( scanner.nextIs( columnDirective ) )
		{
			DalColumnStruct* dalColumnStruct = new DalColumnStruct( _columns.size(), this );
			dalColumnStruct -> readColumnStruct( scanner );
			_columns.push_back( dalColumnStruct );
		}
		else
		{
			expectedOneOf(
				scanner,
				nameDirective.c_str(),
				rowsDirective.c_str(),
				columnDirective.c_str(),
				attributeDirective.c_str(),
				0 );
		} // if
	} // while

	scanner.match( '>' );
	ENDPROC
}


void
DalTableStruct::check( const DataSet*		ds ) const
{
	PROC
	const Block* bk = ds -> block( name() );

	this -> DalAttributableStruct::check( bk );

	if( bk -> type() != Block::TableT )
	{
		errstr << "Expected block with qualifed name '" << bk -> qualifiedName() << "' to be a table" << Sas::error( "wrongBlockType" );
	} // if

	const Table* tb = ds -> table( name() );

	if( columnStructs() != tb -> columns() )
	{
		errstr << "Expected table with qualified name '" << tb -> qualifiedName() << "' to contain " << columnStructs() << " columns but found " << tb -> columns() << Sas::error( "wrongNumberOfColumns" );
	} // if

	for(
		unsigned int i = 0;
		i < columnStructs();
		++i )
	{
		_columns[ i ] -> check( tb );
	} // for
	ENDPROC
}


// *****************************
// DalArrayStruct class beyond here
// *****************************

class DalArrayStruct : public virtual ArrayStruct, public DalBlockStruct
{
	public :
		DalArrayStruct( unsigned int number, DalDataSetStruct* parent );
		~DalArrayStruct();

		Block::Type		type() const;
		Array::DataType		dataType() const;
		bool			dataTypeFound() const;

		void 			readArrayStruct( ScannerT& scanner );
		void			check( const DataSet* dataSet ) const;

	private :
		// Forbid
		DalArrayStruct( const DalArrayStruct& );
		DalArrayStruct& operator=( const DalArrayStruct& );

		Array::DataType		_dataType;
		bool			_dataTypeFound;

};


Block::Type
DalArrayStruct::type() const
{
	PROC
	RETURN Block::ArrayT;
}


DalArrayStruct::DalArrayStruct( unsigned int number, DalDataSetStruct* parent )
	: DalBlockStruct( number, parent ),
	  _dataTypeFound( false )
{
}


DalArrayStruct::~DalArrayStruct()
{
	PROC
	ENDPROC
}


Array::DataType
DalArrayStruct::dataType() const
{
	PROC
	RETURN _dataType;
}


bool
DalArrayStruct::dataTypeFound() const
{
	PROC
	RETURN _dataTypeFound;
}


void
DalArrayStruct::readArrayStruct( ScannerT& scanner )
{
	PROC
	scanner.match( arrayDirective );
	scanner.match( '<' );

	while( !scanner.NextIs( '>' ) && !scanner.EndOfFile() )
	{
		if( scanner.nextIs( nameDirective ))
		{
			readNameStruct( scanner );
		}
		else if( scanner.nextIs( typeDirective ))
		{
			Token	token;
			scanner.SetState( ScannerT::LINE );
			scanner.match( typeDirective );
			/* AI gcc4 porting
			 * error: deprecated conversion from string constant to 'char*'
			 * cast to (char *)
			 */
			token = scanner.MatchOneOf( (char *)"Int8",  (char *)"Uint16",  (char *)"Int16",  (char *)"Uint32", (char *) "Int32", (char *) "Real32",  (char *) "Real64", 0 );
			_dataType = toDalDataType( token.asString() );
			scanner.SetState( ScannerT::NORMAL );
		}
		else if( scanner.nextIs( attributeDirective ))
		{
			readAttributeStruct( scanner );
		}
		else
		{
			expectedOneOf(
				scanner,
				nameDirective.c_str(),
				attributeDirective.c_str(),
				0 );
		} // if
	} // while
	scanner.match( '>' );
	ENDPROC

}


void
DalArrayStruct::check( const DataSet*		ds ) const
{
	PROC

	const Block* bk = ds -> block( name() );

	this -> DalAttributableStruct::check( bk );

	if( bk -> type() != Block::ArrayT )
	{
		errstr << "Expected block with qualified name '" << bk -> qualifiedName() << "' to be an array" << Sas::error( "wrongBlockType" );
	} // if

	const Array* ar = ds -> array( name() );

	if( dataType() != ar -> dataType() )
	{
		errstr << "Expected array with qualified name '" << ar -> qualifiedName() << "' to have data type " << toString( dataType() ) << " but found " << toString( ar -> dataType() ) << Sas::error( "wrongArrayDataType" );
	} // if
	ENDPROC
}


// **********************************
// DalDataSetStruct class beyond here
// **********************************

DalDataSetStruct::DalDataSetStruct( unsigned int number, DataSetStructServer* parent, bool stopIfCheckFails )
	: DalAttributableStruct( number, stopIfCheckFails ),
	  DalChild<DataSetStructServer>( parent )
{
	PROC
	ENDPROC
}


DalDataSetStruct::~DalDataSetStruct()
{
	PROC
	ENDPROC
}


unsigned int
DalDataSetStruct::blockStructs() const
{
	PROC
	RETURN _blocks.size();
}


const BlockStruct*
DalDataSetStruct::blockStruct( unsigned int blockNumber ) const
{
	PROC
	RETURN ( _blocks[ blockNumber ] );
}


const ArrayStruct*
DalDataSetStruct::arrayStruct( unsigned int blockNumber ) const
{
	PROC
	RETURN (DalArrayStruct*)( _blocks[ blockNumber ] );
}


const TableStruct*
DalDataSetStruct::tableStruct( unsigned int blockNumber ) const
{
	PROC
	RETURN (DalTableStruct*)( _blocks[ blockNumber ] );
}


//void
//DalDataSetStruct::check() const
//{
//	checkDataSet( this );
//}

void
DalDataSetStruct::check() const
{
	PROC
	if( !dataSetServer -> exists( name() ) )
	{
		errstr << "Expected a dataset with name '" << name() << "' to exist" << Sas::error( "datasetDoesNotExist" );
	} // if

	DataSet* ds = dataSetServer -> open( name(), Dal::Read );

	this -> DalAttributableStruct::check( ds );

	if( ds -> blocks() != blockStructs() )
	{
		errstr << "Set " << ds -> name() << " contains " << ds -> blocks();
		errstr << " but expected " << blockStructs();
		errstr << " blocks." << Sas::error( "wrongNumberOfBlocks" );
	} // if

	for(
		unsigned int i = 0;
		i < blockStructs();
		++i )
	{
		_blocks[ i ] -> check( ds );
	} // for
	dataSetServer -> close( ds );

	ENDPROC
}


void
DalDataSetStruct::readDataSetStruct( ScannerT& scanner )
{
	PROC
	scanner.match( datasetDirective );
	scanner.match( '<' );
	Token	token;

	while( !scanner.NextIs( '>' ) && !scanner.EndOfFile() )
	{
		if( scanner.nextIs( nameDirective ))
		{
			readNameStruct( scanner );
		}
		else if( scanner.nextIs( attributeDirective))
		{
			readAttributeStruct( scanner );
		}
		else if( scanner.nextIs( tableDirective ) )
		{
			DalTableStruct* dalTableStruct = new DalTableStruct( _blocks.size(), this );
			dalTableStruct -> readTableStruct( scanner );
			_blocks.push_back( dalTableStruct );
		}
		else if( scanner.nextIs( arrayDirective ) )
		{
			DalArrayStruct* dalArrayStruct = new DalArrayStruct( _blocks.size(), this );
			dalArrayStruct -> readArrayStruct( scanner );
			_blocks.push_back( dalArrayStruct );
		}
		else
		{
			expectedOneOf(
				scanner,
				nameDirective.c_str(),
				arrayDirective.c_str(),
				tableDirective.c_str(),
				attributeDirective.c_str(),
				0 );
		} // if
	} // while

	scanner.match( '>' );
	ENDPROC

}


// ****************************************
// DalDataSetStructServer class beyond here
// ****************************************

class DalDataSetStructServer : public virtual DataSetStructServer
{
	public :
		DalDataSetStructServer();
		~DalDataSetStructServer();

		void 			readDataSetStructs( const string& fileName, bool stopIfCheckFails );
		void 			releaseDataSetStructs();
		unsigned int 		dataSetStructs() const;
		const DataSetStruct* 	dataSetStruct( unsigned int dataSetStructNumber ) const;
		DataSetStruct* 		dataSetStruct( unsigned int dataSetStructNumber );

		const DataSetStruct* 	readDataSet( const string& fileName );

	private :
		// Forbid
		DalDataSetStructServer( const DalDataSetStructServer& );
		DalDataSetStructServer& operator=( const DalDataSetStructServer& );

		typedef vector<DataSetStruct*> DataSetStructList;
		DataSetStructList	_dataSetStructs;

};


DalDataSetStructServer::DalDataSetStructServer()
{
	PROC
	ENDPROC
}


DalDataSetStructServer::~DalDataSetStructServer()
{
	PROC
	ENDPROC
}


void
DalDataSetStructServer::readDataSetStructs( const string& fileName, bool stopIfCheckFails )
{
	PROC
	ScannerT	scanner( fileName );

	while( !scanner.EndOfFile() )
	{
		DalDataSetStruct* dataSetStruct = new DalDataSetStruct( _dataSetStructs.size(), this, stopIfCheckFails );
		dataSetStruct -> readDataSetStruct( scanner );
		_dataSetStructs.push_back( dataSetStruct );
	} // if
	ENDPROC
}


unsigned int
DalDataSetStructServer::dataSetStructs() const
{
	PROC
	RETURN _dataSetStructs.size();
}


const DataSetStruct*
DalDataSetStructServer::dataSetStruct( unsigned int dataSetStructNumber ) const
{
	PROC
	RETURN _dataSetStructs[ dataSetStructNumber ];
}


DataSetStruct*
DalDataSetStructServer::dataSetStruct( unsigned int dataSetStructNumber )
{
	PROC
	RETURN _dataSetStructs[ dataSetStructNumber ];
}


void
DalDataSetStructServer::releaseDataSetStructs()
{
	PROC
	ENDPROC
}


const DataSetStruct*
DalDataSetStructServer::readDataSet( const string& setName )
{
	PROC
	DataSet*	dataSet;
	dataSet = dataSetServer -> open( setName, Dal::Read, DataSetServer::Low );
	dataSetServer -> close( dataSet );
	RETURN 0;
}


DataSetStructServer* dataSetStructServer = new DalDataSetStructServer;


